;
; (c) 2018-2020 Vladimir Estivill-Castro
;
;     This program is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program.  If not, see <https://www.gnu.org/licenses/>.
;
;
(DEFUN my-command-line ()
  (OR 
   #+CLISP *args*
   #+SBCL *posix-argv*  
   #+LISPWORKS system:*line-arguments-list*
   #+CMU extensions:*command-line-words*
   nil))

;
;
(defun option-enable (arguments option)
	;(PRINT arguments )
        ( COND  ((null arguments) nil)
                ( (NUMBERP (SEARCH option (car arguments))) T)
          (T (option-enable (cdr arguments) option ) )))
;
;
(defun from-string-to-list (s)
  (let ((L (read-from-string 
           (concatenate 'string "(" s ")"))))
    L))
;
(defun stringToValueForOption (arguments option) ; make a list of those arguments tagged by the option
        ( COND  ((null arguments) '())
                ( (NUMBERP (SEARCH option (car arguments))) 
                                                                (APPEND
                                                                (from-string-to-list (car (CDR arguments))) 
                                                                (stringToValueForOption (cdr arguments) option)
                                                                )
                )
          (T (stringToValueForOption (cdr arguments) option ) )
        ))
;
;
(defun valueForOptionAsString (arguments option) ; make a list of those arguments tagged by the option
        ( COND  ((null arguments) '())
                ( (NUMBERP (SEARCH option (car arguments))) 
                                                                (APPEND
                                                                (list (car (CDR arguments))) 
                                                                (valueForOptionAsString (cdr arguments) option)
                                                                )
                )
          (T (valueForOptionAsString (cdr arguments) option ) )
        ))
;
;
;	Interpreter of Logic-Label Finite-State Machines
;	===============================================
;
;
(IF (option-enable (my-command-line) "-h")
(BLOCK message
        (PRINT "Command line options are:")
        (PRINT "-h       : THIS MESSAGE (c) 2018-2020 Vladimir Estivill-Castro")
        (PRINT "-g       : DEBUG_MODE")
        (PRINT "-k       : KRIPKE_MODE")
        (PRINT "-d       : KRIPKE_DETAIL --- Kripke detail of generation")
        (PRINT "-r       : KRIPKE_RANGE --- Kripke range for integer variables -KRIPKE_RANGE to KRIPKE_RANGE")
        (PRINT "-s       : Add a sensor variable (can be used more than once)")
        (PRINT "-e       : Add an effector variable (can be used more than once)")
        (PRINT "-w       : Add a whiteboard variable (can be used more than once)")
        (PRINT "-f       : Name of NUSMV file, must have -k option")
        (PRINT "-m       : Name of file to read a machine")
        (PRINT "-l       : Limit the number of states, must have -k option")
        (PRINT "-i       : Name of file with initial PC")
        (PRINT "-t       : Name of file with PC to trace")
        (PRINT "-c       : Output the length of lists of states explored ")
        (QUIT)))
;
;
;
;  	DEBUGGING MODE
;
(LET DEBUG_MODE nil)
(SETQ DEBUG_MODE (option-enable (my-command-line) "-g"))
(format *error-output* "DEBUG_MODE ~s~%" (IF DEBUG_MODE "DEBUG_MODE IS ON"  "DEBUG_MODE IS OFF"))
;
;
(LET KRIPKE_MODE nil)
(SETQ KRIPKE_MODE (option-enable (my-command-line) "-k"))
(format *error-output* "KRIPKE_MODE ~s~%" (IF KRIPKE_MODE "KRIPKE_MODE IS ON"  "KRIPKE_MODE IS OFF"))

; default values
(SETQ KRIPKE_DETAIL nil)
(SETQ KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES 50)
(SETQ KRIPKE_RANGE 1)
(SETQ NO_INITIAL_PC T)
(SETQ NO_TRACE_PC T)
(SETQ PRINT_CAPACITY NIL)

(SETQ PRINT_CAPACITY (option-enable (my-command-line) "-c"))

(IF KRIPKE_MODE
	(BLOCK setUp-NUSMV-file


	(SETQ KRIPKE_DETAIL (option-enable (my-command-line) "-d"))
	(print 'KRIPKE_DETAIL)
	(print KRIPKE_DETAIL)

	(IF
		(NOT	(null  (valueForOptionAsString (my-command-line) "-l")  ))
		(BLOCK set-KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES	
			(SETQ KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES (PARSE-INTEGER  (car (valueForOptionAsString (my-command-line) "-l"))))
			(print 'KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES)
			(print KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES)
	        )	
	)

	(IF
		(NOT (NULL   (valueForOptionAsString (my-command-line) "-r") ))
		(BLOCK set-KRIPKE_RANGE	
			(SETQ KRIPKE_RANGE (PARSE-INTEGER  (car (valueForOptionAsString (my-command-line) "-r"))))
			(print 'KRIPKE_RANGE)
			(print KRIPKE_RANGE)
		)	
	)



		(IF (option-enable (my-command-line) "-f")
			(setq filename 
				(concatenate 'string
					(car (valueForOptionAsString (my-command-line) "-f"))
					".nusmv"
				)
			)
			(setq filename "default.nusmv")
		)
		(IF DEBUG_MODE (print filename))
		(defun openToNUSMVFile ()
  			(with-open-file (stream  filename :external-format charset:iso-8859-1
                           	:direction :output
                           	:if-exists :overwrite
                           	:if-does-not-exist :create )
  			(format stream "MODULE main")  (terpri stream) 
 			(terpri stream) 
  			(format stream "VAR")  (terpri stream) 
                        )
		   filename)
		(defun writeToNUSMVFile (content)
  			(with-open-file (stream  filename :external-format charset:iso-8859-1
                           	:direction :output
                           	:if-exists :append
                           	:if-does-not-exist :create )
  			(format stream content)  (terpri stream) )
		   filename)
	)
)
;
;
(defun readLLFSMS ()
	(stringToValueForOption (my-command-line) "-m")	
)
;
(defun readLFSMFromFilename (LLFSMfilename)
	(with-open-file (aMachine LLFSMfilename
                     :direction :INPUT
                     ;:if-does-not-exist (print 'file-for-llfsms-not-found  *error-output*) 
                )
  	(read aMachine) 
	)
)
;
(defun readArrangementOfLFSMFromFilenames (listOfFilenames)
	( COND ((null listOfFilenames) nil)
	    (T  (APPEND (LIST  (readLFSMFromFilename (car listOfFilenames)))
			(readArrangementOfLFSMFromFilenames (cdr listOfFilenames))
		)
	    )

	)
)
;
(IF (option-enable (my-command-line) "-m")
	(BLOCK reading-llfsms
		(SETQ theLLFSMSfiles (readLLFSMS)
		)
		(IF DEBUG_MODE (BLOCK what-files
			(print 'LLFSM-IN_COMAMND_LINE)
			(print theLLFSMSfiles)
		))
		(SETQ TheArragementofLLFSMs (readArrangementOfLFSMFromFilenames theLLFSMSfiles))
	)
)
;
(defun readStatingGlobalPC ()
        (stringToValueForOption (my-command-line) "-i")
)
;
;
(defun readGobalPCformFilename (gobal-pc-filename)
        (with-open-file (aglobalPCfile gobal-pc-filename
                     :direction :INPUT
                     ;:if-does-not-exist (print 'file-for-pc-not-found  *error-output*)
                )
        (read aglobalPCfile)
        )
)
;
(IF (option-enable (my-command-line) "-i")
        (BLOCK reading-initial-globalPC
                (SETQ theGlobalPCfile  (readStatingGlobalPC)
                )
                (IF DEBUG_MODE (BLOCK what-files
                        (print 'INITIAL-GLOBAL-PC-IN_COMMAND_LINE)
                        (print theGlobalPCfile)
                ))
                (SETQ TheStartingGlobalPCFromCommandLine (readGobalPCformFilename  (car theGlobalPCfile)))
		(SETQ NO_INITIAL_PC NIL)
        )
)
;;;;
; options -m -i -t need some refacorint
(IF (option-enable (my-command-line) "-t")
        (BLOCK reading-trace-globalPC
                (SETQ thePCfile  (stringToValueForOption (my-command-line) "-t")
                )
                (IF DEBUG_MODE (BLOCK what-files
                        (print 'TRACE-GLOBAL-PC-IN_COMMAND_LINE)
                        (print thePCfile)
                ))
                (SETQ TheTracingGlobalPCFromCommandLine (readGobalPCformFilename  (car thePCfile)))
                (SETQ NO_TRACE_PC NIL)
        )
)
;


; 	1) STRUCTURE OF A STATE
;	=======================
;
; A LIST of 3 sections
; each section is a LIST of statements
;
; value for a state
(setq state_one 
( LIST
	( LIST '(write-line "S1: On Entry First Statement") '(write-line "S1: On Entry Second Statement") '(SETMDV W1 (PLUS (PLUS Z Z) X)) '(SETMDV U (PLUS W1 Z)) ) 
	( LIST '(write-line "On Exit") '(SETMDV W2 (PLUS U W1)) '(SETMDV W2 (MINUS Y X)) )
	( LIST '(write-line "Internal") '(SETMDV Y (PLUS W2 W1)) '(SETMDV Z (MINUS (PLUS U V) Y))   )
)
)
;
(setq state_zero 
( LIST
	( LIST '(write-line "S0: On Entry First Statement") '(write-line "S0: On Entry Second Statement") '(SETMDV Y 2) '(SETMDV W1 (MINUS X Y)) '(AND Y 0) ) 
	( LIST '(write-line "On Exit") )
	( LIST '(write-line "Internal"))
)
)
;
;
; 	2) EXTRACT THE SECTION: On-Entry, On-exit, Internal
;	===================================================
;
;
(DEFUN extractOnEntry (state)
   (CAR state)
)
(DEFUN extractOnExit (state)
   (CAR (CDR state))
)
(DEFUN extractInternal (state)
   (CAR (CDR (CDR state)))
)
;
;
; testing
(IF DEBUG_MODE (BLOCK debugging
(write "TESTING extract on entry" )
(PRINT (extractOnEntry state_one))
))
;
; 	3) FORMAT OF A CONTEXT
;	======================
;
; a context is a LIST with variable and value alternating.
; a context
(setq simple_context
	(LIST 'X 2 'y 4)
)
; show a state as a triplet
;
(IF DEBUG_MODE (BLOCK debugging
(PRINT state_one)
(PRINT simple_context)
))
;
;
;
; 	3A) RETRIEVE THE VALUE OF VARIABLE FROM A CONTEXT
;
(DEFUN findVariableValue (theVariable context)
   ( cond ( (null context) theVariable)  ; return ungrounded variable if the context is empty, we didn't find it
          ( (EQ (CAR context) theVariable ) (CAR (CDR context) ) ) ; variable found
	  ( t (findVariableValue theVariable (CDR (CDR context) ) ) )
   )
)
; testing
(IF DEBUG_MODE (BLOCK DEBUGGING
(write "TESTING retrieval of values from context" )
(PRINT (findVariableValue 'X simple_context))
(PRINT (CAR simple_context))
(PRINT (findVariableValue 'Y simple_context))
(PRINT (findVariableValue 'Z simple_context))
))
;
;
; 	3B)UPDATE THE VALUE OF VARIABLE IN A CONTEXT
;
;
;
(DEFUN updateVariableValue (theVariable value context)
   ( cond ( (null context) (LIST theVariable value) )  ; return updated context if the context is empty
          ( (EQ (CAR context) theVariable ) 
		(append (LIST theVariable value) (CDR (CDR  context)))
          ) ; variable found
	  (t (append (LIST (CAR context) (CAR (CDR context))) (updateVariableValue theVariable value (CDR (CDR context) ) )))
   )
)
;
(IF DEBUG_MODE (BLOCK DEBUGGING
(write "TESTING updating variables with values in a context" )
(PRINT (updateVariableValue 'X 3 nil))
(setq env_one (updateVariableValue 'X 3 nil))
(PRINT (updateVariableValue 'Y 4 env_one))
(setq env_two (updateVariableValue 'Y 4 env_one))
(PRINT (updateVariableValue 'Y 5 env_two))
))
;
;
; 	3C)IS VARIABLE DEFINED (EXISTS)  IN A CONTEXT
;
;
(DEFUN isVariableDefined (theVariable context)
   ( COND ( (NULL context) nil )  ; return nil if the context is empty
          ( (EQ (CAR context) theVariable ) t
          ) ; variable found
	  (t (isVariableDefined theVariable (CDR (CDR context) ) )) ; search in the rest
   )
)
;
;
(DEFUN is_operator (an_operator)
        ( cond ( (eq  an_operator 'PLUS) t)
               ( (eq  an_operator 'MINUS) t)
               ( (eq  an_operator 'LESS) t)
               ( (eq  an_operator 'AND) t)
               ( (eq  an_operator 'OR) t)
               ( (eq  an_operator 'NOT) t)
               ( (eq  an_operator 'EQUAL) t)
		(t nil)
	)
)
;
; 	4) STRUCTURE OF  STATEMENTS (assignment with SETMDV) and EXPRESSIONS
;	=================================================================
; a)  WRITE-LINE [constant|variable]
; b)  SETMDV  variable simple-expression
;
(IF DEBUG_MODE (block DEBUGGING
 (write "TESTING identifying operators in expressions" )
(PRINT (is_operator 3))
(PRINT (is_operator 'LESS))
(PRINT (is_operator 'PLUS))
(PRINT (is_operator 'MINUS))
(PRINT (is_operator 'AND))
(PRINT (is_operator 'OR))
(PRINT (is_operator 'NOT))
(PRINT (is_operator 'X))
(PRINT (is_operator 'EQUAL))
(write "TESTING identifying operators in statements" )
(write "the third statement" )
(PRINT (CAR (CDR (CDR (extractOnEntry state_one)))))
; the assignment
(PRINT (CAR (CDR ( CDR (CAR (CDR (CDR (extractOnEntry state_one))))))))
; the operator
(PRINT (CAR (CAR (CDR ( CDR (CAR (CDR (CDR (extractOnEntry state_one)))))))))
(PRINT (is_operator (CAR (CAR (CDR ( CDR (CAR (CDR (CDR (extractOnEntry state_one))))))))))
(write "the fourth statement" )
(PRINT (CAR (CDR (CDR (CDR (extractOnEntry state_one))))))
(write "the AND operator " )
(PRINT (CAR (CAR (CDR (CDR (CDR (extractOnEntry state_one)))))))
(PRINT (is_operator (CAR (CAR (CDR (CDR (CDR (extractOnEntry state_one))))))))
))
;
;
;
(DEFUN groundExpression (expression context timeinfo currenttime mode)
       ; debugging (write 'expression) (PRINT expression) (write 'expression)
      (cond  ((numberp  expression) expression)  ; numbers are expressions
	  ( (null expression)  nil)
	  ( (AND (atom expression) (EQUAL T expression)  )  1)
          (  (AND (atom expression) (is_operator expression)) expression )  
          (  (atom expression) (findVariableValue expression context))  
		;  timeinfo should always be the time since entry into the state  => replace by evaluation against current-time
	 ( (EQUAL (CAR expression) 'AFTER) 
	     (IF KRIPKE_MODE   ; in KRIPKE_MODE this is a 1 pr a 0 according to mode
			mode
	    		(LIST 'IF  (LIST '< (groundExpression (CAR (CDR expression)) context timeinfo currenttime mode ) (LIST '- currenttime timeinfo)
	               		   )
		         1 0
                        )   ; we evaluate Boolean expression in {1,0} and not {T,nil}
		)
         )
	  ( t (cons (groundExpression (CAR expression) context timeinfo currenttime mode)   (groundExpression (CDR expression) context timeinfo currenttime mode )))
	)
)
;
(IF DEBUG_MODE (block DEBUGGING
(write "TESTING grounding all variables in simple expression")
(setq firsttime (- (get-universal-time) 10))
(PRINT (groundExpression 'X simple_context  firsttime (get-universal-time)   1)  )
(PRINT (groundExpression 'X simple_context  firsttime (get-universal-time)   0)  )
(PRINT (groundExpression '4 simple_context firsttime (get-universal-time)  0 ))
(PRINT (groundExpression '4 simple_context firsttime (get-universal-time)  1 ))
(PRINT (groundExpression 'Z simple_context firsttime (get-universal-time)  1 ))
(PRINT (groundExpression 'Z simple_context firsttime (get-universal-time)  0 ))
(PRINT (groundExpression '(PLUS X Y) simple_context firsttime (get-universal-time) 0 ))
(PRINT (groundExpression '(PLUS X Y) simple_context firsttime (get-universal-time) 1 ))
(PRINT (groundExpression '(LESS X Y) simple_context firsttime (get-universal-time)  1 ))
(PRINT (groundExpression '(LESS X Y) simple_context firsttime (get-universal-time)  0 ))
(PRINT (groundExpression '(EQUAL X Y) simple_context firsttime (get-universal-time)  0 ))
;
(PRINT (groundExpression '(AFTER X ) simple_context  firsttime (get-universal-time) 0 ))
(PRINT (groundExpression '(AFTER X ) simple_context  firsttime (get-universal-time) 1 ))
(PRINT (groundExpression '(AFTER (PLUS X Y) ) simple_context  firsttime (get-universal-time) 1 ))
(PRINT (groundExpression '(AFTER (PLUS X Y) ) simple_context  firsttime (get-universal-time) 0 ))
(PRINT (groundExpression '(AFTER 100 ) simple_context  firsttime (get-universal-time) 1 ))
(PRINT (groundExpression '(AFTER 100 ) simple_context  firsttime (get-universal-time) 0 ))
))
;
;
(DEFUN groundStatement (statement context timeinfo currenttime mode)
	(IF DEBUG_MODE (PRINT statement))
   ( cond ( (null statement) nil)
	  ( (atom statement) (groundExpression statement context timeinfo currenttime mode))
          ( (EQ (CAR statement) 'WRITE-LINE) statement)
				; ground the third element which is the expression
          ( (EQ (CAR statement) 'SETMDV) (CONS 'SETMDV (cons (CAR (CDR statement))
                                                               (groundExpression  (CDR (CDR statement))  context timeinfo currenttime mode)
                                                             ) 
                                       ) )
				; otherwise it must be a expression
          ( t (groundExpression statement context timeinfo currenttime mode ) )
   )
)
;
;
(IF DEBUG_MODE (block DEBUGGING
(write "TESTING grounding statements")
(PRINT (CAR (extractOnEntry state_one)))
(PRINT (groundStatement (CAR (extractOnEntry state_one)) simple_context  firsttime (get-universal-time) 1))
(PRINT (groundStatement (CAR (extractOnEntry state_one)) simple_context  firsttime (get-universal-time) 0))
(PRINT (CAR (CDR (extractOnEntry state_one))) )
(PRINT (groundStatement (CAR (CDR (extractOnEntry state_one))) simple_context  firsttime (get-universal-time) 1))
(PRINT (groundStatement (CAR (CDR (extractOnEntry state_one))) simple_context  firsttime (get-universal-time) 0))
(PRINT (CAR (CDR (CDR (extractOnEntry state_one))) ))
(PRINT (groundStatement (CAR (CDR (CDR (extractOnEntry state_one)))) simple_context  firsttime (get-universal-time) 1))
(PRINT (groundStatement (CAR (CDR (CDR (extractOnEntry state_one)))) simple_context  firsttime (get-universal-time) 0))
(PRINT (CAR (CDR (CDR (CDR (extractOnEntry state_one))) )))
(PRINT (groundStatement (CAR (CDR (CDR (CDR (extractOnEntry state_one))))) simple_context  firsttime (get-universal-time) 1) )
(PRINT (groundStatement (CAR (CDR (CDR (CDR (extractOnEntry state_one))))) simple_context  firsttime (get-universal-time) 0) )
(PRINT (groundStatement '(LESS X Y) simple_context  firsttime (get-universal-time) 1) )
(PRINT (groundStatement '(LESS X Y) simple_context  firsttime (get-universal-time) 0) )
(PRINT (groundStatement '(EQUAL X Y) simple_context  firsttime (get-universal-time) 0) )
))
;
;
;
(DEFUN restoreOperators ( LIST)
	(cond ( (null LIST) nil)
	      ( (AND (atom LIST) (eq LIST 'PLUS)) '+)
	      ( (AND (atom LIST) (eq LIST 'MINUS)) '-)
	      ( (AND (atom LIST) (eq LIST 'AND)) 'LOGAND)
	      ( (AND (atom LIST) (eq LIST 'OR)) 'LOGIOR)
	      ( (AND (LISTP LIST)  (eq (CAR LIST) 'LESS)        )    ; Boolean operations in {0,1}
			(LIST 'IF (LIST '<  (restoreOperators  (CAR (CDR LIST))) (restoreOperators  (CAR (CDR (CDR LIST)))   ) )         1 0 )
              )
	      ( (AND (LISTP LIST) (eq (CAR LIST) 'EQUAL)       )    ; Boolean operations in {0,1}
			(LIST 'IF (LIST 'EQUAL  (restoreOperators  (CAR (CDR LIST))) (restoreOperators  (CAR (CDR (CDR LIST)))   ) )         1 0 )
              )

	      ( (AND (LISTP LIST) (eq (CAR LIST) 'NOT)) (LIST 'logxor 1 (car (restoreOperators (CDR LIST))))) ; NOT implemented with XOR
		( (LISTP  LIST) (CONS (restoreOperators (CAR LIST)) (restoreOperators (CDR LIST))))
		(T LIST)
	)
)
;
;
;
;
(IF DEBUG_MODE (block DEBUGGING
(write "TESTING restoring operators statements")
(PRINT (CAR (CDR (CDR (extractOnEntry state_one))) ))
(PRINT (restoreOperators  (groundStatement (CAR (CDR (CDR (extractOnEntry state_one)))) simple_context firsttime (get-universal-time)    1)))
(PRINT (restoreOperators  (groundStatement (CAR (CDR (CDR (extractOnEntry state_one)))) simple_context firsttime (get-universal-time)    0)))
(PRINT (restoreOperators  (groundStatement '(LESS X Y) simple_context firsttime (get-universal-time)   1)   ))
(PRINT (restoreOperators  (groundStatement '(LESS X Y) simple_context firsttime (get-universal-time)   0)   ))
(PRINT (restoreOperators  (groundStatement '(EQUAL X Y) simple_context firsttime (get-universal-time)   0)   ))
(PRINT (restoreOperators  (groundStatement '(EQUAL X X) simple_context firsttime (get-universal-time)   0)   ))
(PRINT (restoreOperators  (groundStatement '(EQUAL Y Y) simple_context firsttime (get-universal-time)   0)   ))
(PRINT (restoreOperators  (groundStatement '(NOT 1) simple_context firsttime (get-universal-time)   0)   ))
(PRINT (restoreOperators  (groundStatement '(AFTER X ) simple_context firsttime (get-universal-time)   0)   ))
(PRINT (restoreOperators  (groundStatement '(AFTER X ) simple_context firsttime (get-universal-time)   1)   ))
))
;
;
; 	5) Structure of a composite context
;	==================================
;
;	A LIST of four contexts: the_sensor_variables, the_output_variables, the_whiteboard_variables, the_local_variables
;
;	Keep Sensor variables simple as Boolean values for NUSMV, ie, domain {0,1}
;
(setq sensor_variable_one
	(LIST 'Z 1 'X 0)
)
(setq output_variables_one
	(LIST 'U 1 'V 1)
)
(setq whiteboard_variables_one (LIST 'W1 1 'W2 1))
(setq composite_context_test (LIST sensor_variable_one output_variables_one whiteboard_variables_one simple_context))
;
;
(DEFUN extractContextSensorVariables (compositeContext)
	(nth 0 compositeContext)
)
(DEFUN extractContextOutputVariables (compositeContext)
	(nth 1 compositeContext)
)
(DEFUN extractContextWhiteboardVariables (compositeContext)
	(nth 2 compositeContext)
)
(DEFUN extractContextLocalVariables (compositeContext)
	(nth 3 compositeContext)
)

(IF DEBUG_MODE (block DEBUGGING
(write "TESTING using composite context")
(PRINT (extractContextSensorVariables composite_context_test))
(PRINT (extractContextOutputVariables composite_context_test))
(PRINT (extractContextWhiteboardVariables composite_context_test))
(PRINT (extractContextLocalVariables composite_context_test))
))
;
;
;
;
;	6) Ground statement in composite context
;
;
;   used meaningfully by 'expressionFires" where the AFTER for transitions makes sense
;   used by 'evaluateStatement' with dummy timeinfo and current time as AFTER makes no sense
;   in the statements of sections of states.
;	mode is a FLAF id we are doing this to generate KRIPKE structure and predicates with AFTER evaluate to {1,0}
;
(DEFUN groundStatementInCompositeContext ( statement composite_context  timeinfo currenttime mode)
			(groundStatement
				(groundStatement
					(groundStatement
                      				(groundStatement statement (extractContextLocalVariables composite_context)  timeinfo currenttime mode )
						(extractContextWhiteboardVariables composite_context)  timeinfo currenttime  mode
					)
					(extractContextOutputVariables composite_context)  timeinfo currenttime mode
				)
				(extractContextSensorVariables composite_context)  timeinfo currenttime mode
			)
) 
;
;
; 	7) EVALUATE a statement
;	========================
;
;  evaluate statement once all free variables are grounded, undefined if free variables remain
;
(DEFUN evaluateGroundedStatement(statement )
		(IF DEBUG_MODE (block DEBUGGING
			(PRINT 'in-evaluate-Grounded-Statement)
			(PRINT statement)
		))
	(COND ( (EQ (CAR statement )  'SETMDV) (EVAL (CAR (CDR (CDR statement)))) )
	 (T (EVAL  statement) )
	)
)
;
;
;  evaluate the "statement" in "compositeContext" and return
; updated composite context 
; this function should be used only by the evaluation of sections of states
(DEFUN evaluateStatement(statement compositeContext)
	; if statement is an assignment
	(IF DEBUG_MODE (block DEBUGGING
		(PRINT 'RECEIVED-IN-EVALUATE-STATEMENT)
		(PRINT statement)
		(PRINT compositeContext)
	))

	( cond
            ( (eq (CAR statement) 'SETMDV) ; handle an assignment
		(setq theVariable (CAR (CDR statement))) ; identify variable to update
		(IF DEBUG_MODE (block DEBUGGING
			(PRINT theVariable) ; debugging
			(PRINT 'test-sensor) (PRINT ( isVariableDefined theVariable  (extractContextSensorVariables compositeContext)))
			(PRINT 'test-output) (PRINT ( isVariableDefined theVariable  (extractContextOutputVariables compositeContext)))
			(PRINT 'test-whiteboard) (PRINT ( isVariableDefined theVariable  (extractContextWhiteboardVariables compositeContext)))
			(PRINT 'test-local) (PRINT ( isVariableDefined theVariable  (extractContextLocalVariables compositeContext)))
		));IF DEBUG_MODE (block DEBUGGING
			; grounding variables: first Local, the whiteboard, then output and then sensor
			; and then restore operators PLUS->'+' and MINUS-> '-'
		(setq exprssionAndContext (restoreOperators (groundStatementInCompositeContext  statement compositeContext
					  0 0 0 ; dummy time info
                                          )))
			(IF DEBUG_MODE (BLOCK an-alert
				(PRINT 'GROUNDED-RESTORED-ALL-LEVELS) (PRINT exprssionAndContext)
			))
;			; test if it is external_sensor, then external_effector,  then whiteboard, and finally local
		(cond
			(( isVariableDefined theVariable  (extractContextLocalVariables compositeContext))  
				(IF DEBUG_MODE (BLOCK DEBUGGING
					(PRINT 'is-local) ; update-local context with result
					(PRINT 'value-is) (PRINT  (evaluateGroundedStatement exprssionAndContext) ) ; debugging	
				))
				; build composite context with updated local context
					(append (LIST (extractContextSensorVariables compositeContext))
						(LIST (extractContextOutputVariables compositeContext))
						(LIST (extractContextWhiteboardVariables compositeContext))
						(LIST (updateVariableValue  theVariable   (evaluateGroundedStatement exprssionAndContext) (extractContextLocalVariables compositeContext)))
					)
			); is-local finishes here
			(( isVariableDefined theVariable  (extractContextWhiteboardVariables compositeContext))  
				(IF DEBUG_MODE (BLOCK DEBUGGING
					(PRINT 'is-whiteboard) ; update-whiteboard context with result
					(PRINT 'value-is) (PRINT  (evaluateGroundedStatement exprssionAndContext) ) ; debugging	
				))
				; build composite context with updated whiteboard context
					(append (LIST (extractContextSensorVariables compositeContext))
						(LIST (extractContextOutputVariables compositeContext))
						(LIST (updateVariableValue  theVariable   (evaluateGroundedStatement exprssionAndContext) (extractContextWhiteboardVariables compositeContext)))
						(LIST (extractContextLocalVariables compositeContext))
					)
			); is-whiteboard finishes here
			(( isVariableDefined theVariable  (extractContextOutputVariables compositeContext))  
				(IF DEBUG_MODE (BLOCK DEBUGGING
					(PRINT 'is-output) ; update-output context with result
					(PRINT 'value-is) (PRINT  (evaluateGroundedStatement exprssionAndContext) ) ; debugging	
				))
				; build composite context with updated output context
					(append (LIST (extractContextSensorVariables compositeContext))
						(LIST (updateVariableValue  theVariable   (evaluateGroundedStatement exprssionAndContext) (extractContextOutputVariables compositeContext)))
						(LIST (extractContextWhiteboardVariables compositeContext))
						(LIST (extractContextLocalVariables compositeContext))
					)
			); is-output finishes here
			(( isVariableDefined theVariable  (extractContextSensorVariables compositeContext))  
				(IF DEBUG_MODE (BLOCK DEBUGGING
					(PRINT 'is-sensor) ; update-sensor context with result
					(PRINT 'value-is) (PRINT  (evaluateGroundedStatement exprssionAndContext) ) ; debugging	
				))
				; build composite context with updated sensor context
					(append (LIST (updateVariableValue  theVariable   (evaluateGroundedStatement exprssionAndContext) (extractContextSensorVariables compositeContext)))
						(LIST (extractContextOutputVariables compositeContext))
						(LIST (extractContextWhiteboardVariables compositeContext))
						(LIST (extractContextLocalVariables compositeContext))
					)
			); is-sensor finishes here
			(T  ; not found anywhere, then it is assumed local
					(append (LIST (extractContextSensorVariables compositeContext))
						(LIST (extractContextOutputVariables compositeContext))
						(LIST (extractContextWhiteboardVariables compositeContext))
						(LIST (updateVariableValue  theVariable   (evaluateGroundedStatement exprssionAndContext) (extractContextLocalVariables compositeContext)))
					)
			);
		)
             )
;;
	;; needs to find the context that has the variables, and some may be in different context
	;; so we need to ground them in rounds
	   (t   
		(setq expressionAndContext (restoreOperators (groundStatementInCompositeContext  statement compositeContext
			 0 0 0 ; dumy timeinfo
                 )))
		(IF DEBUG_MODE (BLOCK expression-evaluation
			(PRINT 'expression-And-Context)
			(PRINT expressionAndContext )
                ))
		(eval expressionAndContext)
		compositeContext
	   )
        )
)
;
;
;
; 	8) EVALUATE a section (On-Entry, On-exit, Internal)
;	====================================================
;	====================================================
;
;
;
(DEFUN evaluateSection (section compositeContext) ; return a composite context
	( cond ( (null section) compositeContext) ; a nil section does not affect context
	       ( t	; evaluate the rest of the statements in the context that results from
			; evaluating the first statement
	    	(evaluateSection (CDR section) (evaluateStatement (CAR section) compositeContext))
               )
	)
)
;
;		NOTE: No timeinfo for the sections of states, AFTER makes no sense here

(DEFUN evaluateOnEntry (state compositeContext)
   (evaluateSection  (extractOnEntry state) compositeContext)
)

(DEFUN evaluateOnExit (state compositeContext)
   (evaluateSection  (extractOnExit state) compositeContext )
)

(DEFUN evaluateInternal (state compositeContext)
   (evaluateSection  (extractInternal state) compositeContext )
)

(IF DEBUG_MODE (BLOCK DEBUGGING
(write "TESTING evaluating sections")
(evaluateOnEntry state_one composite_context_test)
(PRINT (evaluateOnEntry state_one composite_context_test))
(PRINT (evaluateOnExit state_one composite_context_test))
(PRINT (evaluateInternal state_one composite_context_test))
))
;
;
;	Definition of LLFSM
;	=========================
;
;	9) An LFFSM is
;		A)  a LIST of states (the first state is the initial state)
;   		b) A LIST of transitions of the from (state-id-for-source, Boolean expression, state-id-for target)
;
(setq machine_one (LIST ( LIST state_zero state_one) (LIST (LIST 0 t 1) (LIST 1 t 1) (LIST 0 nil 0) ) )
)

(IF DEBUG_MODE (PRINT machine_one))
;
;
(DEFUN theStates (machine) ; extract the states of a machine
	(CAR machine) ; the head of the LIST
)

(DEFUN theTransitions (machine) ; extract all of transitions 
	(CAR (CDR machine)) ; the second in the LIST
)

(DEFUN  matchingStateID (transitions stateID) ; extract the sequence of transitions having stateID as a source 
	( COND
		((null transitions) nil)
		( (eq (CAR (CAR transitions)) stateID )   (APPEND (LIST (CAR transitions))  (matchingStateID (CDR transitions) stateID) )
                )
		(t (matchingStateID (CDR transitions) stateID)  )
	)
)

(IF DEBUG_MODE
(BLOCK debugging
(write "TESTING extracting transitions")
(PRINT (matchingStateID (theTransitions  machine_one) 0))
(PRINT 'A-SPACE)
)
)
;
;
;	This evaluates an expression by grounding in a cntex (which would replace the AFTER for the
;	syntactic sugar it represent of comparing the current time against the time in timeour_info (whose value
;	should be the time when first arriving at the state, set in the evaluation of the OnEntry.
;

(DEFUN expressionFires  (boolean_expression  composite_context timeout_info current_time mode)
		(IF DEBUG_MODE (BLOCK debugging-block
			(PRINT 'before-context)
			(PRINT boolean_expression)
			(PRINT composite_context)
			(PRINT timeout_info)
			(PRINT current_time)
		))
	(setq exprssionAndContext (restoreOperators (groundStatementInCompositeContext 
						 boolean_expression composite_context timeout_info current_time mode )))

		(IF DEBUG_MODE (BLOCK debugging-block
			(PRINT 'after-context)
			(PRINT exprssionAndContext)
		))
	(eval exprssionAndContext)
)

(IF DEBUG_MODE (BLOCK DEBUGGING
(write "TESTING transition fires")
(PRINT (expressionFires (CAR (CDR (CAR (matchingStateID (theTransitions  machine_one) 0)))) composite_context_test firsttime (get-universal-time)  0))
(PRINT (expressionFires (CAR (CDR (CAR (matchingStateID (theTransitions  machine_one) 0)))) composite_context_test firsttime (get-universal-time)  1))
;
(PRINT (expressionFires '(AFTER X) composite_context_test firsttime (get-universal-time) 0)  )
(PRINT (expressionFires '(AFTER X) composite_context_test firsttime (get-universal-time) 1)  )
(SETQ secondtime (- (get-universal-time) 5))
(PRINT (expressionFires '(AFTER X) composite_context_test  secondtime (get-universal-time) 0 ))
(PRINT (expressionFires '(AFTER X) composite_context_test  secondtime (get-universal-time) 1 ))
(SETQ thirdtime (- (get-universal-time) 1))
(PRINT (expressionFires '(AFTER X) composite_context_test thirdtime (get-universal-time) 0))
(PRINT (expressionFires '(AFTER X) composite_context_test thirdtime (get-universal-time) 1))
))
;
;

; return whether a transition fired and the
; corresponding target state
;
;
(DEFUN evaluateSequenceOfTransitions (sequenceOfTransitions composite_context timeout_info current_time mode)
	;(SETQ booleanExpressionHasAfter (hasAfter (CAR (CDR (CAR sequenceOfTransitions)))))
	;(print 'booleanExpressionHasAfter)
	;(print booleanExpressionHasAfter)
	( COND
		( (null sequenceOfTransitions)    (LIST nil nil) )
		; expressions evaluated in {0,1} as Boolean expressions, so, it fires if I get a number not-equal to zero 
		( (AND (NUMBERP
			(expressionFires (CAR (CDR (CAR sequenceOfTransitions))) composite_context   timeout_info current_time mode)
			) 
                   (/= (expressionFires (CAR (CDR (CAR sequenceOfTransitions))) composite_context timeout_info current_time   mode ) 0)
		  )
			(print 'expression-Fires)
                     (LIST T (CAR (CDR ( CDR (CAR sequenceOfTransitions) )))   )
		  
		)
		(t  			(IF DEBUG_MODE (PRINT sequenceOfTransitions) )
			(evaluateSequenceOfTransitions (CDR sequenceOfTransitions) composite_context timeout_info current_time mode ))
	)
)
;
;
;
;

(IF DEBUG_MODE (BLOCK DEBUGGING
(write "TESTING evaluating transitions")
(PRINT (evaluateSequenceOfTransitions (matchingStateID (theTransitions  machine_one) 0) composite_context_test firsttime (get-universal-time) 0))
(PRINT (evaluateSequenceOfTransitions (matchingStateID (theTransitions  machine_one) 0) composite_context_test firsttime (get-universal-time) 1))
))
;
;   running a ringlet on a machine from its current state and whether last ringlet fired a transition, the third parameter is time since entry into the state
;   result is a triplet, the composite-context, the new (or same state id), whether it fired, and maybe a new time
(DEFUN runRinglet (machine compositeContext current_state has_fired timeout_info mode)
		(IF DEBUG_MODE (BLOCK testing
			(PRINT (LIST 'RECEIVED-RUN-RINGLET-CURRENT-STATE current_state
			(IF (>= current_state 0 ) (nth current_state (theStates machine)))))
		))
	(if has_fired    ; if we are arriving to a state we record the time we get here (get-universal-time) in the time-info  
	  (BLOCK we_just_arrive
              (setq newtime (get-universal-time) )
	      (setq first_context (evaluateOnEntry (nth current_state (theStates machine) ) compositeContext) )	
	  )
	  (setq first_context  compositeContext) 	
	)
        ; select transitions out of the current state
	(setq sequenceOfTransitions (matchingStateID (theTransitions machine) current_state))
	;; if there are no transitions, the state is final
	( COND
		( (OR (null sequenceOfTransitions   ) (< current_state 0))  ; state has not transitions out or machine decomissioned with negative state ID
			(PRINT 'MACHINE-INEFECTIVE)
			(LIST first_context -1 nil timeout_info)
		) ; case for state being final
		( T   ; otherwise
			(setq resultTransitions (evaluateSequenceOfTransitions sequenceOfTransitions 
                                                               first_context 
                                                               (IF has_fired newtime timeout_info) ;use the new time if we executed the OnEntry
                                                              (get-universal-time)
								mode
                         ))  ; returns if transition fired and target state or not firing, it requires a time snapshot to evaluate AFTER predicates

			(IF DEBUG_MODE (BLOCK DEBUGGING (PRINT 'resultTransitions) (PRINT resultTransitions)))
			(COND
			( (CAR resultTransitions) 
	    		 	(IF DEBUG_MODE (PRINT 'transition-fired) )
			; any transition that fires causes OnExit to run instead of Internal
				(setq exit_context (evaluateOnExit (nth current_state (theStates machine) ) first_context) )

	    			(IF DEBUG_MODE (BLOCK debugOnExitResult  (PRINT 'after-on-exit) (PRINT exit_context) ))
				; return resulting context, next state, if next state is different than current state, then the transition is signaled
				; as fired so that next ringlet run OnEntry and the timeout-info is reset in the onEntry							
				(LIST  exit_context  (CAR (CDR resultTransitions)) (NOT (EQ current_state (CAR (CDR resultTransitions)))) timeout_info ) 
			)
	    		( T   
				(IF DEBUG_MODE  (PRINT 'no-transition-fired))
				(setq exit_context (evaluateInternal (nth current_state (theStates machine) ) first_context) )
				(LIST  exit_context  current_state nil (IF has_fired newtime timeout_info)  ) ; return resulting context, next state, transition fired, and new time if we updated
	    		)
			); end COND
		)
	) ; COND check if there are transitions
)

;
(IF DEBUG_MODE (BLOCK DEBUGGING
(WRITE "TESTING evaluating ringlets")
(PRINT (runRinglet  machine_one  composite_context_test 0 T (get-universal-time)  1 ))
(PRINT (runRinglet  machine_one  composite_context_test 0 T (get-universal-time)  0 ))
))
;
;
;	10) An ARRANGEMENT of LLFSMs is LIST of LLFSMs
;
;
(setq state_light_on 
( LIST
	( LIST '(write-line "light on: on entry")  '(SETMDV W2 (PLUS (MINUS W1 Z) X)) '(AND X 0) ) 
	( LIST '(write-line "light on: on exit") '(SETMDV W2 (PLUS X W2)) '(SETMDV Y (MINUS X W2)) )
	( LIST '(write-line "light on: internal") '(SETMDV Y (PLUS W2 W1)) '(SETMDV Z (MINUS (PLUS U V) Y))   )
)
)

(setq state_light_off 
( LIST
	( LIST '(write-line "light off: on entry")  '(SETMDV V (MINUS (PLUS W1 Z) Y)) '(OR X 0) ) 
	( LIST '(write-line "light off: on exit") '(SETMDV W1 (PLUS X Y)) '(SETMDV W2 (MINUS X Y)) )
	( LIST '(write-line "light off: internal") '(SETMDV U (PLUS W2 W1)) '(SETMDV V (MINUS (PLUS W1 W2) Y))   )
)
)

(setq machine_two (LIST ( LIST state_light_on state_light_off) (LIST (LIST 0 '(AND X W1) 1) (LIST 1 '(OR (PLUS U V) W1) 0) (LIST 0 '(AND W1 W2) 0) ) )
)

(IF DEBUG_MODE (BLOCK DEBUGGING
(PRINT 'machine-two)
(PRINT machine_two)
))

(IF (NOT (option-enable (my-command-line) "-m"))
	(setq TheArragementofLLFSMs (LIST machine_one machine_two))
)

;
;
;
;	EXECUTION
;	===============
;	10) Executing  an ARRANGEMENT of LLFSMs is running a ringlet of each
;	LLFSM in round-robin fashion
;	10A) initially all local context are empty, so is the whiteboard
;
;	10b) The global program counter is
;	10b-i) a turn that indicates which LLFSM is the next turn
;	10b-ii) a composite context with three contexts
;		(sensor-variables, effector variables, whiteboard variables) 
;	10b-iii) a LIST of 4-tuple that are the pcs for each LLFSM in the arrangement
;	 (the local variables [i.e. the local context of this LLFSM], current_state, whether fired, info on timeout-predicates)
;	10b-iii-I; information for timeout-predicates (AFTER expression) is an integer, the time-value snapshot when we first arrived at this state.
;		the vlaue is the universal time recorded on the first evaluation of the OnEntry
;	
;	10c) The input is
;	10c-i) A context of sensor variables; variables and values
;	10c-ii) A LIST of output variables (no values)
;	10c-iii) A LIST of whiteboard variables (no values)
;
(DEFUN extractCurrentMachine ( machine-id arrangement )
	(nth machine-id arrangement)
)

(DEFUN extractSensorVariablesFromPC (global_pc) ; from the second  of global_pc; take first
	(CAR (CAR (CDR global_pc)))
)

(DEFUN extractEffectorVariablesFromPC (global_pc) ; from the second  of global_pc; take second
	(CAR (CDR (CAR (CDR global_pc))))
)

(DEFUN extractWhiteboardVariablesFromPC (global_pc) ; from the second  of global_pc; take third
	(CAR (CDR (CDR (CAR (CDR global_pc)))))
)

(DEFUN extractListOfContextsFromPC (global_pc) ; take third from global_pc
	(CAR  (CDR  (CDR global_pc)))
)

(DEFUN extractLocalContextCurrentMachine ( machine-id global_pc ) ; from the (third of global_pc) take nth and then first
       (CAR 	(nth machine-id  (CAR (CDR  (CDR global_pc  )))) )
)

(DEFUN extractCurrentStateOfCurrentMachine ( machine-id global_pc )  ; from the (third of global_pc) take nth and then second
       (CAR  (CDR	(nth machine-id  (CAR (CDR  (CDR global_pc  )))) ))
)

;
(DEFUN extractWhetherFiredCurrentMachine ( machine-id global_pc ) ; from the (third of global_pc) take nth and then third
       (CAR  (CDR (CDR	(nth machine-id  (CAR (CDR  (CDR global_pc  )))) )))
)
;
;
(DEFUN extractInfoOnTimeoutPredicateForCurrentMachine ( machine-id global_pc) ; from the (third of global_pc) take the nth and then fourth
       (CAR (CDR  (CDR (CDR	(nth machine-id  (CAR (CDR  (CDR global_pc  )))) ))))
)
;



;
(DEFUN initialize_empty (LIST_variables)
	(COND
	  ( (null LIST_variables)	 nil)
	   (T  (append	 (LIST (CAR LIST_variables) 0) (initialize_empty (CDR LIST_variables))  ))
	)
)

(IF DEBUG_MODE (BLOCK DEBUGGING
(write "TESTING: building empty initialization for LIST of variables")
(PRINT (initialize_empty '(A B C D) ))
))

;
;   For interactive reading of sensor variables
;
;
(DEFUN retrieveNumber (variableName)
   (loop
	(print 'PROVIDE-NEW-VALUE)
   	(print variableName)
   	(setq answer (read))
   	(when (numberp answer) (return (list variableName answer) ))
   )
)


(DEFUN readSensorExternalVariables (simple-context)
	( COND ((null simple-context ) nil)
	  (t 
         	(print (LIST 'CURRENT-SENSOR-VARIABLE-VALUE-PAIR (list (CAR simple-context) (CAR (CDR simple-context)))))
                (APPEND
		   (retrieveNumber (car simple-context) )
			(readSensorExternalVariables (CDR (CDR simple-context)))
		)
		)	
	)
)

;;;;;;;;;;;;;;
; 	Produce all combinations of a context of binary valued variables (only for  Kripke structure of external variables)
(defun populateContextAllBinary (context)
        (COND
                ((null context) nil)
                ((null (cdr (cdr context))) (list (list (car context) 1) (list (car context) 0)  ))
                ( T     
                        (APPEND
                                (mapcar #'(lambda (x)  (APPEND  (list ( car context)) (APPEND  '(1) x))) (populateContextAllBinary (cdr (cdr context))))
                                (mapcar #'(lambda (x)  (APPEND  (list ( car context)) (APPEND  '(0) x))) (populateContextAllBinary (cdr (cdr context))))
                        )
                )
        )
)

;;
(defun buildListoGobalContextFromListOfExternalContext (initial listSensorContext listEffectorContext  currentMachineIDcounter lengthOfArrangement theEnvironment theListOfLocalEnvironments)
	(IF DEBUG_MODE 
			(BLOCK details
				(print 'BUILD-LIST-OF-GOBAL-CONTEXT-FROM-LIST-OF-EXTERNAL-CONTEXT)
				(print 'BUILD-LIST-OF-GOBAL-CONTEXT-FROM-LIST-OF-EXTERNAL-CONTEXT)
				(print 'parameters)
				(print 'initial) (print initial)
				(print 'listSensorContext) (print listSensorContext)
				(print 'listEffectorContext) (print listEffectorContext)
				(print 'currentMachineIDcounter) (print currentMachineIDcounter)
				(print 'lengthOfArrangement) (print lengthOfArrangement)
				(print 'theEnvironment) (print theEnvironment)
				(print 'theListOfLocalEnvironments) (print theListOfLocalEnvironments)
				(print 'end-parameters)
	))

	(COND
		( (EQ 1 (length listEffectorContext)	)
			(SETQ theNeWEnvironment 
				(LIST   (extractContextSensorVariables  theEnvironment) (CAR listEffectorContext)  (extractContextWhiteboardVariables theEnvironment)
				   (extractContextLocalVariables theEnvironment)
				)
		        )
			(buildListoGobalContextFromListOfSensorContext initial listSensorContext currentMachineIDcounter lengthOfArrangement theNeWEnvironment theListOfLocalEnvironments)
                )
		( T
			(SETQ theNeWEnvironment 
				(LIST   (extractContextSensorVariables  theEnvironment) (CAR listEffectorContext)  (extractContextWhiteboardVariables theEnvironment)
				   (extractContextLocalVariables theEnvironment)
				)
		        )
			(APPEND
				(buildListoGobalContextFromListOfSensorContext initial listSensorContext currentMachineIDcounter lengthOfArrangement theNeWEnvironment theListOfLocalEnvironments)
				(buildListoGobalContextFromListOfExternalContext  initial listSensorContext (CDR listEffectorContext)  currentMachineIDcounter lengthOfArrangement theEnvironment theListOfLocalEnvironments)
			)
		)
	)
)
;
;	expand a global_pc by all possible sub-Kripke state of variations of external sensor variables
;	if is the initial expansion, do not increment the turn of whose machine is it
;
(defun buildListoGobalContextFromListOfSensorContext (initial listSensorContext  currentMachineIDcounter lengthOfArrangement theEnvironment theListOfLocalEnvironments)
        (COND 
                (       (null listSensorContext) Nil )
                ( T
                    (CONS
                        (LIST (IF initial 0 (mod (+ currentMachineIDcounter  1) lengthOfArrangement  )) ;the turn
                              (LIST (car listSensorContext)   (extractContextOutputVariables  theEnvironment) (extractContextWhiteboardVariables theEnvironment))
                                theListOfLocalEnvironments
                        )
                        (buildListoGobalContextFromListOfSensorContext initial (cdr listSensorContext) currentMachineIDcounter lengthOfArrangement theEnvironment theListOfLocalEnvironments   )
                    )
                )
        )
)

(defun updateLocalEnvironment (list-local-environments new-local-environment position)
		(COND
			((EQ position 0)  (CONS new-local-environment (CDR list-local-environments)))
			(T
			   (CONS (CAR list-local-environments)	(updateLocalEnvironment (CDR list-local-environments) new-local-environment (- position 1) ))
			)
		)
)

(defun buildNextGlobalPCFromOutcome (arrangement currentMachineID old-PC outcome)
		(setq newEnvironment (CAR outcome)) ; extract the new environment
                (IF DEBUG_MODE (BLOCK DEBUGGING
                        (print (LIST 'OUTCOME outcome))
                        (print (LIST 'NEWENVIRONMENT newEnvironment))
                        (PRINT (extractContextLocalVariables newEnvironment))
                ))
	   ; extract new local PC
          (setq newLocalPC (LIST (extractContextLocalVariables newEnvironment)  (CAR (CDR outcome)) (CAR (CDR (CDR outcome ))) (CAR (CDR (CDR (CDR outcome ))))       )) 
                (IF DEBUG_MODE (BLOCK DEBUGGING
                        (print (LIST 'NEWLOCALPC newLocalPC))
                ))   
          (SETQ List-local-environments (extractListOfContextsFromPC old-PC))
          		(IF DEBUG_MODE (BLOCK DEBUGGING
                                (print (LIST 'LIST-LOCAL-ENVIRONMENTS List-local-environments))
          			(PRINT (CAR (CDR (CDR global_pc))))
          			(PRINT (LIST 'WHO-WE-REPLACE (nth currentMachineID (CAR (CDR (CDR global_pc)))) ))
          		))
	   (SETQ  List-local-environments (updateLocalEnvironment List-local-environments newLocalPC  currentMachineID ))
                (IF DEBUG_MODE (BLOCK DEBUGGING
                        (print (LIST 'LIST-LOCAL-ENVIRONMENTS List-local-environments))
                        (print (LIST 'OLD-GLOBAL_PC global_pc))
                ))
          ; update sensor
          (IF (NOT KRIPKE_MODE) (BLOCK interactive
                (print (QUOTE "***> INTERACTIVE-MODE <***"))
                (print (LIST 'OUTPUT/EFFECTOR-VALUES (extractContextOutputVariables newEnvironment)))
                (print 'WHITEBOARD-VALUES)
		(print (extractContextWhiteboardVariables newEnvironment))
                (print (LIST 'CURRENT-MACHINE-ID  currentMachineID 'CURRENT-STATE-ID_IN_MACHINE (extractCurrentStateOfCurrentMachine currentMachineID global_pc )  ))
                (print (LIST 'LOCAL-VALUES (extractContextLocalVariables  newEnvironment)))
		;(loop for aLFSM in arrangement	do (PRINT aLFSM)) ; for each machien print name and local variables/values
                (SETQ theNewSensors (readSensorExternalVariables (extractContextSensorVariables newEnvironment) ) )
                        ; reset the global pc
                (SETQ  global_pc (LIST (mod (+ currentMachineID 1) (LIST-length arrangement )  )
                     (LIST theNewSensors  (extractContextOutputVariables  newEnvironment) (extractContextWhiteboardVariables newEnvironment))
                         List-local-environments
                        ))
                (IF DEBUG_MODE (BLOCK DEBUGGING
                    (print (LIST 'UPDATED-GLOBAL_PC  global_pc)  *error-output*)
                ))
                (list global_pc)
                )
                ;else  list of Kripke states are all combinations of external (sensor) variables
                     (buildListoGobalContextFromListOfSensorContext nil   (populateContextAllBinary  (extractContextSensorVariables newEnvironment))
                                currentMachineID
                                (LIST-LENGTH  arrangement)
                                newEnvironment
                         List-local-environments
                      )
          );closes if
)

(defun  nextGlobalPCs (arrangement global_pc)
           (setQ currentMachineID (CAR global_pc) )  ; which machine has its turn
                (IF DEBUG_MODE (BLOCK DEBUGGING
                        (PRINT (LIST 'EXTRACT-CURRENT-MACHINE (extractCurrentMachine currentMachineID   arrangement)))
                        (PRINT (LIST 'EXTRACT-LOCAL-CONTEXT-CURRENT-MACHINE (extractLocalContextCurrentMachine currentMachineID  global_pc )))
                        (PRINT (LIST 'EXTRACT-CURRENT-STATE-OF-CURRENT-MACHINE (extractCurrentStateOfCurrentMachine currentMachineID global_pc )))
                        (PRINT (LIST 'EXTRACT-WHETHER-CURRENT-MACHINE-FIRED (extractWhetherFiredCurrentMachine currentMachineID global_pc )))
                        (PRINT (LIST 'EXTRACT-INFO-ON-TIMEOUT-PREDICATE-FOR-CURRENT-MACHINE (extractInfoOnTimeoutPredicateForCurrentMachine currentMachineID global_pc )))
                ))

           (setq outcome (runRinglet  (extractCurrentMachine currentMachineID  arrangement) ; run a ringlet in the machine obtain new context, whether it fired, and timeoutinfo
                        (LIST
                                (extractSensorVariablesFromPC global_pc)
                                (extractEffectorVariablesFromPC global_pc)
                                (extractWhiteboardVariablesFromPC global_pc)
                                (extractLocalContextCurrentMachine currentMachineID global_pc ) ; the composite context
                        )
                        (extractCurrentStateOfCurrentMachine currentMachineID  global_pc )      ; the current state of current machine
                        (extractWhetherFiredCurrentMachine currentMachineID global_pc ) ; whether fired
			(extractInfoOnTimeoutPredicateForCurrentMachine currentMachineID global_pc ) ; info on timer
			0
                ) ; runRinglet
           )

      	(IF (NOT KRIPKE_MODE)
	            (buildNextGlobalPCFromOutcome arrangement currentMachineID global_pc outcome) 
		; else
		(APPEND (buildNextGlobalPCFromOutcome arrangement currentMachineID global_pc outcome) 
			(buildNextGlobalPCFromOutcome arrangement currentMachineID global_pc 
				(runRinglet  (extractCurrentMachine currentMachineID  arrangement) ; run a ringlet in the machine obtain new context, whether it fired, and timeoutinfo
                         	(LIST
                                	(extractSensorVariablesFromPC global_pc)
                                	(extractEffectorVariablesFromPC global_pc)
                                	(extractWhiteboardVariablesFromPC global_pc)
                                	(extractLocalContextCurrentMachine currentMachineID global_pc ) ; the composite context
                        	)
                        		(extractCurrentStateOfCurrentMachine currentMachineID  global_pc )      ; the current state of current machine
                        		(extractWhetherFiredCurrentMachine currentMachineID global_pc ) ; whether fired
                        		(extractInfoOnTimeoutPredicateForCurrentMachine currentMachineID global_pc ) ; info on timer
                        	1
                                ) ; runRinglet	
			) ; second buildNextGlobalPCFromOutcome
		) ; append
	); closes the if
)

(DEFUN build_initial_global_pc (arrangement sensor_variables effector_variables whiteboard)
	(SETQ sensor_context (initialize_empty sensor_variables ))
	(SETQ output_context (initialize_empty effector_variables ))
	(SETQ whiteboard_shared_context (initialize_empty whiteboard ))
	(SETQ local-environments (MAKE-LIST (LENGTH arrangement)))
	(FILL local-environments (LIST '() 0 T (get-universal-time)  )) ; every Local environment is initially empty,
						; every machine has state 0 as its initial state
						; we arrive at this state for the first time, the initial state executes its on-entry
						; the information on time-out is (current-time), so the time-out is not active
		(IF DEBUG_MODE (BLOCK DEBUGGING
			(PRINT 'LOCAL-ENVIRONMENTS)
			(PRINT local-environments)
		))

        ; Make all environments canonical; i.e. they variables are sorted alphabetically
        ; that way, at least nothing not local to a machine is already sorted
                (SETQ theSensorVariablesSorted (backFromCanonical (canonicalContext sensor_context )))
                (SETQ themotorVariablesSorted (backFromCanonical (canonicalContext  output_context )))
                (SETQ thewhiteboardVariablesSorted (backFromCanonical (canonicalContext  whiteboard_shared_context)))

	(SETQ  global_pc (LIST 0 (LIST theSensorVariablesSorted  themotorVariablesSorted thewhiteboardVariablesSorted ) 
			local-environments
        ) )

                (SETQ theContextforLocalvariables  (extractListOfContextsFromPC global_pc) )

	(IF DEBUG_MODE (BLOCK show-local
        	(print 'THE-CONTEXT-FOR-LOCAL-VARIABLES)
        	(print theContextforLocalvariables)
	))
	(print 'INITIAL-GLOBAL_PC  *error-output*)
	(print global_pc *error-output*)
)

;;
;;	convert a context into a list of pairs : (variable-name value) to
;;	sort on the variable name
;
(DEFUN pairsContext ( context)
        (COND ((null context)  nil)
                (T
                        (CONS (list ( car context) (car (cdr context)) ) (pairsContext  (cdr ( cdr context))))
                )
        )
)
;
(DEFUN canonicalContext (context)
(sort   (pairsContext context)  #'(lambda (x y) (string-lessp (car x) (car Y))))
)
;
(DEFUN backFromCanonical (canonical-context)
	(COND
		( (NULL canonical-context) nil)
		( T 	(CONS  (car (car canonical-context))
				(CONS  (car (cdr (car canonical-context)))
					(backFromCanonical (cdr canonical-context))
				)
			)
		)
	)
)
;
(DEFUN isPairedContextEqual ( context-a context-b)
        ( COND
                (   (AND (null context-a) (null context-b))  T) ; two nill contxt are equal
                (   (AND (null context-a) (NOT (null context-b)))  nil) ; second nil context only, then not equal
                (   (AND (NOT (null context-a)) (null context-b))  nil) ; first  nil context only, then not equal
                (   (NOT (AND (EQ (car (car context-a)) (car (car context-b)))  ; text the variable
                         (EQ (car (cdr (car context-a))) (car ( cdr (car context-b))));text the value
                    ))    nil
                )
                (T (isPairedContextEqual (cdr context-a) (cdr context-b)))
        )
)
;
;
;;;;;
;
;
; only the conditions in common, not each initial state
(defun printInitialKripkeStates (theWhiteboard  ThePCsofMachines )
		      (writeToNUSMVFile "INIT")  ; the things in common of all initial states for sure are
					           ; turn is 0, whiteboard variables set to zero
						   ;	all machines are in state 0, with fired on/true so they execute their EnEntry
			
			( writeToNUSMVFile
				(concatenate 'string "turn = 0 & " theWhiteboard ThePCsofMachines
				)	
			)
)
;
;	for printing in NUSMV file the type creates a prefix to variables
(defun toStringContext (aConetxt contextType variableNumber)
	(COND    ((null aConetxt) "")
		(T     	( concatenate 'string
			 (format nil  "~A~D~A = ~A & "  contextType variableNumber  (car aConetxt) (car (cdr aConetxt))  )
			  (toStringContext (cdr (cdr aConetxt)) contextType (+ 1 variableNumber))
			)
		)
	)
)
;
(defun toStringContextNextKripkeState (aConetxt contextType variableNumber)
	(COND    ((null aConetxt) "")
		(T     	( concatenate 'string
			 (format nil  "next(~A~D~A) = ~A & "  contextType variableNumber  (car aConetxt) (car (cdr aConetxt))  )
			  (toStringContextNextKripkeState (cdr (cdr aConetxt)) contextType (+ 1 variableNumber))
			)
		)
	)
)
;
;
(defun toStringMachinePC (machinePC  machineNumber) 
		( concatenate 'string
		  (toStringContext (car machinePC) "M" machineNumber   )
		  (format nil  "M~Dpc = ~D & M~Dfired = ~D" machineNumber (car (cdr machinePC))  machineNumber (IF (car (cdr (cdr machinePC))) 1 0)   )
		)
)
;
(defun toStringTheMachinesKripkeState ( theMachinesListOfPCs number)
	( COND ( (null theMachinesListOfPCs) "")
		((EQ 1 ( length theMachinesListOfPCs)) (toStringMachinePC (car theMachinesListOfPCs) number))
		( T
			( concatenate 'string
			   (toStringMachinePC (car theMachinesListOfPCs) number )
			   " & "
			   (toStringTheMachinesKripkeState (cdr theMachinesListOfPCs) (+ 1 number))
			)
		)
	)
)
;
(defun toStringMachinePCasNextKripke (machinePC  machineNumber)
                ( concatenate 'string
                  (toStringContextNextKripkeState (car machinePC) "M" machineNumber   )
                  (format nil  "next(M~Dpc) = ~D & next(M~Dfired) = ~D" machineNumber (car (cdr machinePC))  machineNumber (IF (car (cdr (cdr machinePC))) 1 0)   )
                )
)
;
(defun toStringTheMachinesAsNextKripkeState ( machinePCs number)
	( COND ( (null machinePCs) "")
		((EQ 1 ( length machinePCs)) (concatenate 'string  (toStringMachinePCasNextKripke (car machinePCs) number) " " ) ; transition with out CASE finishes with no semicolon in NUSMV
                )
		( T
			( concatenate 'string
			   (toStringMachinePCasNextKripke (car machinePCs) number )
			   " & "
			   (toStringTheMachinesAsNextKripkeState (cdr machinePCs) (+ 1 number))
			)
		)
	)
)
;

;
(defun toStringKripkeState (a-global-pc-to-print  is_it_source)
	(concatenate 'string
		(format nil "( turn = ~A & "  (car a-global-pc-to-print))
		(toStringContext (extractSensorVariablesFromPC a-global-pc-to-print) "E" 0  )
		(toStringContext (extractEffectorVariablesFromPC a-global-pc-to-print) "E" (- (length (extractSensorVariablesFromPC a-global-pc-to-print)) 2)  )
		(toStringContext (extractWhiteboardVariablesFromPC a-global-pc-to-print) "W" 0  )
		(toStringTheMachinesKripkeState (extractListOfContextsFromPC a-global-pc-to-print) 0)
		;(IF is_it_source ":")   ; the source of a transition has s colon
		(IF is_it_source " & ")   ; the source of a transition has s colon
	)
)

(defun toStringNextKripkeState (a-global-pc)
	(concatenate 'string
		(format nil "        next(turn) = ~A & "  (car a-global-pc))
		(toStringContextNextKripkeState (extractSensorVariablesFromPC a-global-pc) "E" 0  )
		(toStringContextNextKripkeState (extractEffectorVariablesFromPC a-global-pc) "E" (- (length (extractSensorVariablesFromPC a-global-pc)) 2)  )
		(toStringContextNextKripkeState (extractWhiteboardVariablesFromPC a-global-pc) "W" 0  )
		(toStringTheMachinesAsNextKripkeState (extractListOfContextsFromPC a-global-pc) 0)
	)
)

;
(defun isLocalConetxtEqual (a-local-pc b-local-pc)
		(COND
			( (NOT (EQ  (CAR (CDR a-local-pc)) (CAR (CDR b-local-pc))  ) ) nil )                    ; test if they have the same PC; i.e.e second element
			( (NOT (EQ  (CAR (CDR (CDR a-local-pc))) (CAR  (CDR (CDR b-local-pc)))  ) ) nil )       ; test if they have the same fired condition; i.e. third element
		        (  (NOT	(isPairedContextEqual
					(canonicalContext (car a-local-pc))
					(canonicalContext (car b-local-pc))
				))   nil )							; the local variables and values are not equal
			(T 
				;(print '(canonicalContext (car a-local-pc))) (print (canonicalContext (car a-local-pc))) (print '(canonicalContext (car b-local-pc))) (print (canonicalContext (car b-local-pc)))
				T)
		)	
)
;
(defun areListOfLocalContextsEqual ( a-local-list b-local-list)
		( COND
			( (AND (NULL a-local-list)  (NULL b-local-list) )  T)  ; two empty list are equal, return true escape recursion
			( (AND (NOT (NULL a-local-list))  (NULL b-local-list) )  nil)  ; return false if one is empty and the other not
			( (AND  (NULL a-local-list)  (NOT (NULL b-local-list) ) )  nil)  ; return false if one is empty and the other not
			(  (NOT (isLocalConetxtEqual (CAR a-local-list) (CAR b-local-list)))  nil   ; if the PC for the head machine are not equal return nil
			) 
			( T   (areListOfLocalContextsEqual (cdr a-local-list) (cdr b-local-list)))
		)
)
;
;
(defun areListOfLocalContextsEqualBIS ( a-local-list b-local-list)
		( COND
			( (AND (NULL a-local-list)  (NULL b-local-list) )  T)  ; two empty list are equal, return true escape recursion
			( (AND (NOT (NULL a-local-list))  (NULL b-local-list) )  nil)  ; return false if one is empty and the other not
			( (AND  (NULL a-local-list)  (NOT (NULL b-local-list) ) )  nil)  ; return false if one is empty and the other not
			(  (NOT (isLocalConetxtEqual (CAR a-local-list) (CAR b-local-list)))  nil   ; if the PC for the head machine are not equal return nil
			) 
			( T 
				(print 'A-LOCAL-LIST)
				(print a-local-list)
				(print 'B-LOCAL-LIST)
				(print b-local-list)
			  (areListOfLocalContextsEqualBIS (cdr a-local-list) (cdr b-local-list)))
		)
)
;
(defun areKripkeStatesEqual (a-global-pc b-global-pc)
	(IF DEBUG_MODE (BLOCK who-are-we-comparing
		(print 'ARE-Kripke-STATES-EQUAL)
		(print a-global-pc )
		(print b-global-pc)
	))
	( COND (  (NOT (EQ (CAR a-global-pc) (car b-global-pc)))
			  (IF DEBUG_MODE (print 'turn-not-equal))
                   nil   
         )  ; not same turn => not equal
		( (NOT (isPairedContextEqual  
			(pairsContext (extractSensorVariablesFromPC a-global-pc))
			(pairsContext (extractSensorVariablesFromPC b-global-pc))
			))    
			  (IF DEBUG_MODE (print 'external-sensor-variables-not-equal) )
                        nil 
                )							; not the same values for external sensor variables => not equal
		( (NOT (isPairedContextEqual  
			(pairsContext (extractEffectorVariablesFromPC a-global-pc))
			(pairsContext (extractEffectorVariablesFromPC b-global-pc))
			))  
			  (IF DEBUG_MODE (print 'external-effector-variables-not-equal))
                          nil 
                )							; not the same values for external effector variables => not equal
		( (NOT (isPairedContextEqual  
			(pairsContext (extractWhiteboardVariablesFromPC a-global-pc))
			(pairsContext (extractWhiteboardVariablesFromPC b-global-pc))
			))    
			  (IF DEBUG_MODE (print 'whiteboard-variables-not-equal))
                        nil 
                )							; not the same values for external effector variables => not equal
		;check  local context
		(  (NOT ( areListOfLocalContextsEqual (extractListOfContextsFromPC  a-global-pc)   (extractListOfContextsFromPC  b-global-pc)))  
			  (IF DEBUG_MODE (print 'list-of-local-PC-context-not-equal))
                   nil
		)
	     ( T	;(print (extractListOfContextsFromPC  a-global-pc)) (print (extractListOfContextsFromPC  b-global-pc)) 
			;( areListOfLocalContextsEqualBIS (extractListOfContextsFromPC  a-global-pc)   (extractListOfContextsFromPC  b-global-pc))
              T)				; no difference found
	)
)
;
;
;	find if the given Kripke state is in the list
(defun DisInListOfPCs (aGlobaPC ListOfKripkeStates)
	;(IF DEBUG_MODE (BLOCK searchingPCs
		(print 'IS-IN-LIST-OF-PCS)
		(print aGlobaPC)
		(print ListOfKripkeStates)
	;))
	(COND  ( (null ListOfKripkeStates) nil )
		( (areKripkeStatesEqual aGlobaPC (car ListOfKripkeStates)) 
		                      (IF DEBUG_MODE (BLOCK PCinfo	
						(print 'FOUND) (print (car ListOfKripkeStates)) 
					))
                     T      )
		( T  (isInListOfPCs aGlobaPC (cdr ListOfKripkeStates)) )
	)
)


;
;	find if the given Kripke state is in the list
(defun isInListOfPCs (aGlobaPC ListOfKripkeStates)
	(IF DEBUG_MODE (BLOCK searchingPCs
		(print 'IS-IN-LIST-OF-PCS)
		(print aGlobaPC)
		(print ListOfKripkeStates)
	))
	(COND  ( (null ListOfKripkeStates) nil )
		( (areKripkeStatesEqual aGlobaPC (car ListOfKripkeStates)) 
		                      (IF DEBUG_MODE (BLOCK PCinfo	
						(print 'FOUND) (print (car ListOfKripkeStates)) 
					))
                     T      )
		( T  (isInListOfPCs aGlobaPC (cdr ListOfKripkeStates)) )
	)
)

(IF DEBUG_MODE (BLOCK toTestPCmatching
	(SETQ PC-ONE
		'(1 ((BUTTONPUSHED 0 DOOROPEN 0) (LIGHT 1 MOTOR 0 SOUND 1) (TIMELEFT 0))
 ((NIL 0 NIL 3758971319) (NIL 0 T 3758971319) (NIL 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(SETQ IS-LIST_WITHPC-TWO
		'((2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
  ((NIL 0 NIL 1000) (NIL 0 NIL 1000) ((CURRENTTIME 0) 1 NIL 1000) (NIL 0 T 1000))))
	)
	(print PC-ONE)
	(print IS-LIST_WITHPC-TWO)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 0) (LIGHT 1 MOTOR 0 SOUND 1) (TIMELEFT 0))
 ((NIL 0 NIL 3758971319) (NIL 0 T 3758971319) (NIL 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 1 MOTOR 0 SOUND 1) (TIMELEFT 0))
 ((NIL 0 NIL 3758971319) (NIL 0 T 3758971319) (NIL 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 0))
 ((NIL 0 NIL 3758971319) (NIL 0 T 3758971319) (NIL 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 T 3758971319) (NIL 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) (NIL 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) ((CURRENTTIME 1) 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	; combinations
	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) ((CURRENTTIME 1) 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))


	; fix current time first
	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) ((CURRENTTIME 0) 0 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) ((CURRENTTIME 0) 1 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) ((CURRENTTIME 0) 1 NIL 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	; fix the local pc
	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) ((CURRENTTIME 1) 1 T 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))

	(SETQ PC-ONE
		'(2 ((BUTTONPUSHED 0 DOOROPEN 1) (LIGHT 0 MOTOR 0 SOUND 1) (TIMELEFT 1))
 ((NIL 0 NIL 3758971319) (NIL 0 NIL 3758971319) ((CURRENTTIME 1) 1 NIL 3758971319)
  (NIL 0 T 3758971319)))
	)
	(print PC-ONE)
	(IF (isInListOfPCs PC-ONE IS-LIST_WITHPC-TWO) (print 'FOUND) (print 'NOT-FOUND))


))



;
(defun remove-duplicateOutcome ( list-outcomesPCs)
	(COND
		( (EQ 1 (length list-outcomesPCs))  list-outcomesPCs )
		((isInListOfPCs (car  list-outcomesPCs) (CDR list-outcomesPCs) )  (remove-duplicateOutcome  (CDR list-outcomesPCs) ) )
                (T (CONS (car  list-outcomesPCs) (remove-duplicateOutcome  (CDR list-outcomesPCs) ))) 
	)
)
;
;(print 'testing-here)
;	expand a Kripke state in the list to expand
;	move it to the expanded list if not already there
;	execute a ringlet, and check children if known (expended or to exapnd)
;	if Kripke structure is being generated, add unknown to kripke_states_pending_to_expand
;	if Kripke is not being generated, add the child to kripke_states_pending_to_expand
(defun runFromListToExpand (arrangement   kripke_states_pending_to_expand expanded_kripke_states )
          (IF DEBUG_MODE
			(print '********************************* )
			(print 'RUN-FROM-LIST-TO-EXPAND )
	  )
	(setQ statesGenerated 1)
	(setQ debugginGenerated 1)
	(loop
		(setQ statesGenerated (+ statesGenerated 1))
		(setq expanding_global_pc (car kripke_states_pending_to_expand))   ; extract a global_pc
		(setq kripke_states_pending_to_expand (cdr kripke_states_pending_to_expand))
		(IF KRIPKE_DETAIL (BLOCK infoOnQueueManagement
			(print (LIST 'STATES-GENERATED statesGenerated   ))
			(print (LIST 'LENGTH-OF-PENDING-STATES (length kripke_states_pending_to_expand)  ))
			(print (LIST 'HEAD-OF-kripke_states_pending_to_expand (CAR kripke_states_pending_to_expand)))
			(IF DEBUG_MODE (print kripke_states_pending_to_expand)  )
			;(print kripke_states_pending_to_expand)  
			(print (LIST 'LENGTH-OF-EXPANDED-STATES (length expanded_kripke_states)))
			(print (LIST 'HEAD-OH-expanded_kripke_states (CAR expanded_kripke_states)))
			(IF DEBUG_MODE  (print expanded_kripke_states))
			;(print expanded_kripke_states)
			(print '****************************************)
			(print 'EXPANDING)
			(print 'EXTRACTED-FRONT-GLOBAL-PC )
			(print expanding_global_pc)
		))
		(IF PRINT_CAPACITY (BLOCK reflectmemory-footprint
			(PRINT (LIST 'GENERATED_STATES statesGenerated)   *error-output*)
			(PRINT (LIST 'LENGTH-OF-PENDING-STATES (length kripke_states_pending_to_expand) )  *error-output*)
			(PRINT (LIST 'LENGTH-OF-EXPANDED-STATES (length expanded_kripke_states))   *error-output*)
		))

		(IF NO_TRACE_PC (PRINT 'no-tracing)
			(BLOCK a-trace-state-is-found
			(IF  (DisInListOfPCs expanding_global_pc (LIST TheTracingGlobalPCFromCommandLine ) )   
				(BLOCK traced-pc-found
					(print 'TRACED_PC_FOUND)
					(print 'statesGenerated)
					(print statesGenerated)
					(print expanding_global_pc)
			          (QUIT)
				)
                        ) 
		))

                (setQ  outcomeFromOneRinglet (nextGlobalPCs  arrangement expanding_global_pc))

		(IF KRIPKE_DETAIL (BLOCK infoOnExapnsionResult
			(print '***********************************)
			(print 'OUTCOME-FROM-ONE-RINGLET)
			(print (length outcomeFromOneRinglet))
			(print outcomeFromOneRinglet)
		))

		;; we may have duplicate outputs, so we need to eliminate them
		(IF (LENGTH outcomeFromOneRinglet) (SETQ outcomeFromOneRinglet  (remove-duplicateOutcome  outcomeFromOneRinglet) ))

               (IF KRIPKE_DETAIL (BLOCK infoOnExapnsionResult
                        (print '***********************************)
                        (print 'UNIQUE-OUTCOME-FROM-ONE-RINGLET)
                        (print (length outcomeFromOneRinglet))
                        (print outcomeFromOneRinglet)
                ))

      		(IF (NOT KRIPKE_MODE)
		; all unknown states states add them
                       		(SETQ kripke_states_pending_to_expand outcomeFromOneRinglet)  ; only one PC expanded, must not be a KRIPKE outcome
				; else we need to check what we generated
				(BLOCK  dealing-with-the-list-of-next-states
					; for each outcome state, we need to print the Kripke transition
					; and we need to test whether we seen it before
					; is in anywhere in our two lists we DO NOT insert it
					; if its unseen, it goes on the first list
					(IF KRIPKE_DETAIL (BLOCK whatToExpand
						(PRINT 'SOURC-KRIPKE-STATE	)
						(PRINT expanding_global_pc	)
					))
					(loop
					 	(SETQ childKripkeState (CAR outcomeFromOneRinglet)) 
					 	(SETQ outcomeFromOneRinglet (CDR outcomeFromOneRinglet))
						(IF KRIPKE_DETAIL (BLOCK source
							(PRINT 'CHILD-KRIPKE-STATE	)
							(PRINT childKripkeState	)
						))
						; global_pc is the source
					        (writeToNUSMVFile (toStringKripkeState expanding_global_pc  T ))
					        (writeToNUSMVFile (toStringNextKripkeState childKripkeState))
					        (writeToNUSMVFile ")")

						;(if (> debugginGenerated 5) (quit) (SETQ debugginGenerated (+ 1 debugginGenerated))  )

						(IF (NOT (OR
								(isInListOfPCs childKripkeState expanded_kripke_states)
								(isInListOfPCs childKripkeState kripke_states_pending_to_expand)
							))
					  	(SETQ kripke_states_pending_to_expand (CONS childKripkeState    kripke_states_pending_to_expand))
						 (IF KRIPKE_DETAIL (print 'NOT-NEW-STATE))
						); if not in either list
						(IF (NOT (isInListOfPCs childKripkeState expanded_kripke_states))
					  	     (SETQ expanded_kripke_states (CONS childKripkeState    expanded_kripke_states))
						)

						; only print the disjunction separator if not at the end of everything
						(IF (NOT (= (length kripke_states_pending_to_expand) 0 ))
					        	(writeToNUSMVFile "|")
							(BLOCK elsePartShouldBeTheLastTarnsition
								(IF (NOT  (null outcomeFromOneRinglet) )
					        				(writeToNUSMVFile "|")
										(writeToNUSMVFile " ")
								)
							)
						)

						(WHEN (null outcomeFromOneRinglet) (RETURN NIL))
					); loop of outcome
				); if NOT KRIPKE_MODE
		)
		(WHEN (AND KRIPKE_MODE (> statesGenerated KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES)) (RETURN NIL)) ; stop if Kripke mode
		(WHEN (AND KRIPKE_MODE (= (length kripke_states_pending_to_expand) 0 )) (RETURN NIL)) ; stop if Kripke mode
	); outer loop

	(IF (AND KRIPKE_MODE (> statesGenerated KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES))
		(BLOCK early-termination
			(print 'KRIPKE_MAXIMUM_NUMBER_EXPANDED_STATES-REACHED  *error-output*)
			(print (LIST 'LENGTH-OF-PENDING-STATES (length kripke_states_pending_to_expand)  ))
			(print (LIST 'LENGTH-OF-EXPANDED-STATES (length expanded_kripke_states)))
		)
	)
	(IF (AND KRIPKE_MODE (= (length kripke_states_pending_to_expand) 0 ))
			(print 'COMPLETE_HRIPKE_MODEL-BUILT  *error-output*)
	)
		
)
;
;(print 'testing-here-2)
;
(defun writeNUSMBooleanVariable (contextType variableNumber variable)
	(IF DEBUG_MODE (BLOCK info-on-boolean-variable
		(print 'VARIABLE-NUMBER)
		(print variableNumber)
	))
	 (writeToNUSMVFile (format nil "~A~D~A:{0,1};" contextType variableNumber variable))	
)
;
;
(defun writeNUSMContextOfBooleanVariable (contextType context number)
	(IF DEBUG_MODE (BLOCK info-on-boolean-variable
		(print 'NUSMV-CONTEXT)
		(print context)
	))
	( COND
		( (null context)  nil)
		(T  (BLOCK write-varibales
			(writeNUSMBooleanVariable contextType number (car context))
			(writeNUSMContextOfBooleanVariable contextType   (cdr context) (+ number 1) )
		    )
		)
	)
)
;
(defun writeNUSMMachinesFired (arrangement number)
	(COND
		( (null arrangement)  nil)
		( T (BLOCK print-list-fired
			(writeToNUSMVFile 
		             (format nil  "M~Dfired:{0,1};" number)
			)
			(writeNUSMMachinesFired (cdr arrangement) (+ 1 number))
		    )
		)
	)
)
;
(defun writeNUSMMachinesPC (arrangement number)
	(COND
		( (null arrangement)  nil)
		( T (BLOCK print-list-fired
			(writeToNUSMVFile 
		             (format nil  "M~Dpc:-1 .. ~D;" number   (- (length (theStates (car arrangement)) ) 1) )
			)
			(writeNUSMMachinesPC (cdr arrangement) (+ 1 number))
		    )
		)
	)
)


; find Left-hand Sides of assignments as candidates for Local variables
; from sections of a state of an LLFSM, which is a list of statements
(defun findCandidatesInSection (stateSection)
	(COND
		( ( null stateSection) nil)
		( T     (APPEND ; if the first atom in a statement is SETMDV then next atom is a candidate
				(IF (EQ (CAR (CAR stateSection))   'SETMDV)  (LIST (CAR (CDR  (CAR stateSection))))
					nil
				)
				(findCandidatesInSection (CDR stateSection))
			)
		)
	)
)

; find Left-hand Sides of assignments as candidates for Local variables
; from their states of an LLFSM
(defun findCandidates (theMachineStates)
	(COND
		( ( null theMachineStates) nil)
		( T
			(APPEND (findCandidatesInSection (extractOnEntry (car theMachineStates)))
			        (findCandidatesInSection (extractOnExit (car theMachineStates)))
			        (findCandidatesInSection (extractInternal (car theMachineStates)))
				(findCandidates (CDR theMachineStates))
			)
		)
	)
)

(IF DEBUG_MODE (BLOCK debugging
(write "TESTING extract candidate local variables" )
(print (findCandidates (theStates machine_two )  ) )
(print (remove-duplicates (findCandidates (theStates machine_two )  ) ))
))

(defun writeTestedCandidatesAsLocalVariables (CandidateLocalVariables theCompositeContext machineNumber)
	(COND
		( (null CandidateLocalVariables)  nil)
		( T 
			(IF (NOT (OR
				 (isVariableDefined  (CAR CandidateLocalVariables) (extractContextSensorVariables theCompositeContext)) 
				 (isVariableDefined  (CAR CandidateLocalVariables) (extractContextOutputVariables theCompositeContext)) 
				 (isVariableDefined  (CAR CandidateLocalVariables) (extractContextWhiteboardVariables theCompositeContext)) 
			    ))	
                            (writeToNUSMVFile (format nil "M~D~A: - ~D .. ~D;" machineNumber (CAR CandidateLocalVariables) KRIPKE_RANGE KRIPKE_RANGE  ))
			)
		        (writeTestedCandidatesAsLocalVariables (CDR CandidateLocalVariables ) theCompositeContext machineNumber  )
		)
	)
)


(IF (AND KRIPKE_MODE DEBUG_MODE) (BLOCK debugging
(SETQ candidateLocalVariables (remove-duplicates (findCandidates (theStates machine_two )  ) ))
(writeTestedCandidatesAsLocalVariables candidateLocalVariables composite_context_test   0)
))

(defun writeNUSMVMachineLocalVariables (arrangement theCompositeContextForTest machineNumber)
	(COND
		( (null arrangement)  nil)
		( T ( BLOCK print-list-local-variables
			(print  (remove-duplicates (findCandidates (theStates (car arrangement)))))
			(writeTestedCandidatesAsLocalVariables
				(remove-duplicates (findCandidates (theStates (car arrangement))))
				theCompositeContextForTest machineNumber
			)
		    	(writeNUSMVMachineLocalVariables (cdr arrangement) theCompositeContextForTest (+ 1 machineNumber))
		))
	)
)


(defun compareVariableNames (varNameOne varNameTwo)
	(string-lessp 
	  (format nil "~A" varNameOne)
	  (format nil "~A" varNameTwo)
	)
)


(defun sortNamesVariables (list-variables)
	(sort list-variables ( lambda (x y ) (compareVariableNames x y) ) )
)

;
(DEFUN generate_execution_for_arrangement  (arrangement  sensor_context effector_variables whiteboard )

	;; TEST FROM SPECIFIC PC
	(IF  NO_INITIAL_PC
	   (SETQ global_pc (build_initial_global_pc  arrangement  sensor_context effector_variables whiteboard ))
	   (SETQ global_pc TheStartingGlobalPCFromCommandLine )
	)
	
	(print global_pc)
	(IF DEBUG_MODE 
		(print 'AFTER-BUILD-INITIAL-GLOBAL-PC)
	)

	(IF KRIPKE_MODE 
		(BLOCK Kripke-setup
			(print 'GENERATING-KRIPKE-STRUCTURE *error-output*)
			(openToNUSMVFile  )

			(IF NO_INITIAL_PC (BLOCK expanding-all-externals
        					(setQ currentMachineID (CAR global_pc) )  ; which machine has its turn
						; expand initially all external variables, so all possible values are Kripke states
						; both sensor and effector variables are expanded
						(SETQ kripke_states_pending_to_expand
		     					(buildListoGobalContextFromListOfExternalContext T  
					        		(populateContextAllBinary  (extractSensorVariablesFromPC global_pc))
					        		(populateContextAllBinary  (extractEffectorVariablesFromPC global_pc))
								currentMachineID
								(LIST-LENGTH  arrangement)
								(car (cdr global_pc))
								(extractListOfContextsFromPC global_pc)
		     					)
						)
					) ;; else initial state provided, use only that one
					(SETQ kripke_states_pending_to_expand (list global_pc))
			  )
			  (IF DEBUG_MODE (BLOCK kripke_states_pending_to_expand
					(PRINT '***************************)
					(PRINT 'KRIPKE_STATES_PENDING_TO_EXPAND )
					(PRINT kripke_states_pending_to_expand)
			  ))
			
			(SETQ turnVariable
				(IF (EQ 1  (LENGTH arrangement )) "turn : 0..0;"
					(concatenate 'string  "turn : 0.." (format nil "~A" (- (LENGTH arrangement) 1)  ) ";"
					)
				)
			)
 			(writeToNUSMVFile turnVariable) ; writes the run
			; write variables of sensor context starting from zero
			(writeNUSMContextOfBooleanVariable "E" (sortNamesVariables sensor_context) 0)
			(writeNUSMContextOfBooleanVariable "E" (sortNamesVariables effector_variables) (length sensor_context))
			(writeNUSMContextOfBooleanVariable "W" (sortNamesVariables whiteboard) 0)
			; the Boolean on each machine PC
			(writeNUSMMachinesFired arrangement 0)
			(writeNUSMMachinesPC arrangement 0)

			; write local variables as INTs
			(writeNUSMVMachineLocalVariables arrangement (CAR (cdr global_pc)) 0)			

			; writing the initial states
 			(writeToNUSMVFile "")
			(printInitialKripkeStates 
				(toStringContext (extractWhiteboardVariablesFromPC global_pc) "W" 0  )
				(toStringTheMachinesKripkeState (extractListOfContextsFromPC global_pc) 0)
                        )
 			(writeToNUSMVFile "TRANS")
			; We believe no combination of CASE with TRANS
 			;(writeToNUSMVFile "case")
		) ; ELSE to  BLOCK Kripke setup
		(BLOCK run-one-deterministic-execution-path
			(print 'generating-one-path-execution *error-output*)
			(SETQ kripke_states_pending_to_expand (list global_pc))
		); execution case
	)

		; execute until pending to expand is empty, the current list of pending states is the list of known states
	(runFromListToExpand arrangement  kripke_states_pending_to_expand kripke_states_pending_to_expand ) 
	;in interactive mode we do not reach beyond here
 	;(writeToNUSMVFile "TRUE    : next(turn) = turn;")
 	;(writeToNUSMVFile "esac")
)
;
;(print 'testing-here-3)
;
(setq theSensorVariables
(IF  (option-enable (my-command-line) "-s")
        (BLOCK   extractListOfvariables  ;(print 'command-line-sensor-variables *error-output*)
		(stringToValueForOption (my-command-line) "-s")
        ) 
	(BLOCK defualtSensorvariables (print 'junk) (print 'default-sensor-variables *error-output*) '(Z X)
	)
)
)
;
(setq theEffectorVariables
(IF  (option-enable (my-command-line) "-e")
        (BLOCK   extractListOfvariables (stringToValueForOption (my-command-line) "-e")
        )
	(BLOCK defualtSensorvariables (print 'default-effector-variables *error-output*) '(U V)
	)
)
)
;
(setq theWhiteboardVariables
(IF  (option-enable (my-command-line) "-w")
        (BLOCK   extractListOfvariables (stringToValueForOption (my-command-line) "-w")
        )
	(BLOCK defualtSensorvariables (print 'default-whiteboard-variables *error-output*) '(W2 W1)
	)
)
)
;
(format T " ")
(format T "(c) 2018-2020 Vladimir Estivill-Castro")
(format T  " ")
(print 'THE-ARRAGEMENT-OF-LLFSMS)
(print TheArragementofLLFSMs)
(print 'MACHINE-ZERO)
(print (car TheArragementofLLFSMs))
(print 'THE-STATES)
(print (theStates (car TheArragementofLLFSMs)))
(print 'THE-TRANSITIONS)
(print (theTransitions (car TheArragementofLLFSMs)))
;
;run
(generate_execution_for_arrangement  TheArragementofLLFSMs  theSensorVariables  theEffectorVariables theWhiteboardVariables  )     


